@@ -1,6 +1,6 @@
 /* otfview.c -- View glyphs of OpenType fonts.
 
-Copyright (C) 2003, 2004, 2005, 2006, 2008
+Copyright (C) 2003, 2004, 2005, 2006, 2008, 2009
   National Institute of Advanced Industrial Science and Technology (AIST)
   Registration Number H15PRO167
 
@@ -47,6 +47,19 @@ write to the Free Software Foundation, Inc., 59 Temple Place, Suite
 
 #include <otf.h>
 
+#define CAST_FROM_XTPOINTER(TYPE, DATA, VAR)	\
+  do {						\
+    long TYPE temp = (long TYPE) (DATA);	\
+    (VAR) = temp;				\
+  } while (0)
+
+#define XtAddCallbackWithCast(TYPE, W, PROC, VAR)		\
+  do {								\
+    long TYPE temp = (long TYPE) (VAR);				\
+    XtAddCallback (W, XtNcallback, PROC, (XtPointer) temp);	\
+  } while (0)
+
+
 #define DEFAULT_PIXEL_SIZE 30
 int pixel_size;
 
@@ -72,6 +85,9 @@ XtAppContext context;
    | | idxh[7] glyph[112]  ...    glyph[127]| |
    | |         idxl[0]     ...    idxl[15]  | |
    | +--------------------------------------+ |
+   | +---- uvs_area (box) (optional) -------+ |
+   | | uvs[?].w ...                         | |
+   | +--------------------------------------+ |
    | +--- render_area (form) ---------------+ |
    | | clear del bidi alt_subst             | |
    | | +--- raw (box) --------------------+ | |
@@ -88,6 +104,7 @@ Widget shell, frame;
 Widget command_area, quit, dump, *charmap;
 Widget navi_area, FIRST, PREV, prev, range, next, NEXT, LAST;
 Widget glyph_area, glyph[128], index_label[8];
+Widget uvs_area, uvs_label;
 Widget render_area, clear, del, bidi, alt_subst, raw, seq;
 Widget raw_label, raw_image, seq_label, seq_image;
 unsigned long foreground, background;
@@ -154,6 +171,13 @@ struct {
   int codes[64];
 } glyph_rec;
 
+OTF_EncodingSubtable14 *sub14 = NULL;
+
+struct {
+  Widget w;
+  int c;
+} uvs[256];
+
 OTF *otf;
 char *filename;
 int fontindex;
@@ -242,6 +266,27 @@ update_glyph_area ()
   XtSetValues (range, arg, 1);
 }
 
+void
+update_uvs_area (int c)
+{
+  OTF_GlyphID code[256];
+  Arg arg[1];
+  int i;
+
+  OTF_get_variation_glyphs (otf, c, code);
+
+  for (i = 0; i < 256; i++)
+    if (uvs[i].w)
+      {
+	if (code[i])
+	  XtSetArg (arg[0], XtNsensitive, True);
+	else
+	  XtSetArg (arg[0], XtNsensitive, False);
+	XtSetValues (uvs[i].w, arg, 1);
+      }
+}
+
+
 char *
 get_features (OTF_FeatureList *list, FeatureRec *rec)
 {
@@ -313,7 +358,11 @@ update_seq_area ()
   gstring.glyphs = malloc (sizeof (OTF_Glyph) * len);
   memset (gstring.glyphs, 0, sizeof (OTF_Glyph) * len);
   for (i = 0; i < len; i++)
-    gstring.glyphs[i].c = gstring.glyphs[i].glyph_id = glyph_rec.glyphs[i];
+    {
+      gstring.glyphs[i].c = glyph_rec.codes[i];
+      if (charmap_index < 0)
+	gstring.glyphs[i].glyph_id = glyph_rec.glyphs[i];
+    }
 
   XFillRectangle (display, seq_pixmap, gc, 0, 0, render_width, render_height);
   XDrawLine (display, seq_pixmap, gc_set, 0, glyph_y, render_width, glyph_y);
@@ -322,6 +371,7 @@ update_seq_area ()
       char *str;
 
       OTF_drive_gdef (otf, &gstring);
+      OTF_drive_cmap (otf, &gstring);
       if (otf->gsub)
 	{
 	  str = get_features (&otf->gsub->FeatureList, &gsub);
@@ -339,7 +389,7 @@ update_seq_area ()
 	  str = get_features (&otf->gpos->FeatureList, &gpos);
 	  if (str)
 	    {
-	      OTF_drive_gpos (otf, &gstring, NULL, NULL, str);
+	      OTF_drive_gpos2 (otf, &gstring, NULL, NULL, str);
 	      free (str);
 	    }
 	}
@@ -375,69 +425,78 @@ update_seq_area ()
       int prev_width;
       int advance = bmp->advance;
 
-      if (! bmp->pixmap)
+      if (gstring.glyphs[i].glyph_id && ! bmp->pixmap)
 	{
 	  create_pixmap (gstring.glyphs[i].glyph_id);
 	  if (! bmp->pixmap)
 	    continue;
 	  advance = bmp->advance;
 	}
-      switch (g->positioning_type)
+      if (g->positioning_type)
 	{
-	case 0:
-	  break;
+	  while (1)
+	    {
+	      switch (g->positioning_type)
+		{
+		case 1: case 2:
+		  {
+		    int format = g->f.f1.format;
 
-	case 1: case 2:
-	  {
-	    int format = g->f.f1.format;
+		    if (format & OTF_XPlacement)
+		      xoff = g->f.f1.value->XPlacement * pixel_size / unitsPerEm;
+		    if (format & OTF_XPlaDevice)
+		      xoff += DEVICE_DELTA (g->f.f1.value->XPlaDevice, pixel_size);
+		    if (format & OTF_YPlacement)
+		      yoff = g->f.f1.value->YPlacement * pixel_size / unitsPerEm;
+		    if (format & OTF_YPlaDevice)
+		      yoff += DEVICE_DELTA (g->f.f1.value->YPlaDevice, pixel_size);
+		    if (format & OTF_XAdvance)
+		      advance += g->f.f1.value->XAdvance * pixel_size / unitsPerEm;
+		    if (format & OTF_XAdvDevice)
+		      advance += DEVICE_DELTA (g->f.f1.value->XAdvDevice,
+					       pixel_size);
+		  }
+		  break;
 
-	    if (format & OTF_XPlacement)
-	      xoff = g->f.f1.value->XPlacement * pixel_size / unitsPerEm;
-	    if (format & OTF_XPlaDevice)
-	      xoff += DEVICE_DELTA (g->f.f1.value->XPlaDevice, pixel_size);
-	    if (format & OTF_YPlacement)
-	      yoff = g->f.f1.value->YPlacement * pixel_size / unitsPerEm;
-	    if (format & OTF_YPlaDevice)
-	      yoff += DEVICE_DELTA (g->f.f1.value->YPlaDevice, pixel_size);
-	    if (format & OTF_XAdvance)
-	      advance += g->f.f1.value->XAdvance * pixel_size / unitsPerEm;
-	    if (format & OTF_XAdvDevice)
-	      advance += DEVICE_DELTA (g->f.f1.value->XAdvDevice, pixel_size);
-	  }
-	  break;
+		case 3:
+		  /* Not yet supported.  */
+		  break;
+		case 4: case 5:
+		  if (! base)
+		    break;
+		  prev = base;
+		  prev_width = base_width;
+		  goto label_adjust_anchor;
+		default: 		/* i.e. case 6 */
+		  if (! mark)
+		    break;
+		  prev = mark;
+		  prev_width = 0;
+		label_adjust_anchor:
+		  {
+		    int base_x, base_y, mark_x, mark_y;
 
-	case 3:
-	  /* Not yet supported.  */
-	  break;
-	case 4: case 5:
-	  if (! base)
-	    break;
-	  prev = base;
-	  prev_width = base_width;
-	  goto label_adjust_anchor;
-	default: 		/* i.e. case 6 */
-	  if (! mark)
-	    break;
-	  prev = mark;
-	  prev_width = 0;
-	label_adjust_anchor:
-	  {
-	    int base_x, base_y, mark_x, mark_y;
+		    base_x = g->f.f4.base_anchor->XCoordinate * pixel_size / unitsPerEm;
+		    base_y = g->f.f4.base_anchor->YCoordinate * pixel_size / unitsPerEm;
+		    mark_x = g->f.f4.mark_anchor->XCoordinate * pixel_size / unitsPerEm;
+		    mark_y = g->f.f4.mark_anchor->YCoordinate * pixel_size / unitsPerEm;
 
-	    base_x = g->f.f4.base_anchor->XCoordinate * pixel_size / unitsPerEm;
-	    base_y = g->f.f4.base_anchor->YCoordinate * pixel_size / unitsPerEm;
-	    mark_x = g->f.f4.mark_anchor->XCoordinate * pixel_size / unitsPerEm;
-	    mark_y = g->f.f4.mark_anchor->YCoordinate * pixel_size / unitsPerEm;
-
-	    if (g->f.f4.base_anchor->AnchorFormat != 1)
-	      adjust_anchor (g->f.f4.base_anchor, face, prev, &base_x, &base_y);
-	    if (g->f.f4.mark_anchor->AnchorFormat != 1)
-	      adjust_anchor (g->f.f4.mark_anchor, face, g, &mark_x, &mark_y);
-	    xoff = (base_x - prev_width) - mark_x;
-	    yoff = base_y - mark_y;
-	  }
+		    if (g->f.f4.base_anchor->AnchorFormat != 1)
+		      adjust_anchor (g->f.f4.base_anchor, face, prev, &base_x, &base_y);
+		    if (g->f.f4.mark_anchor->AnchorFormat != 1)
+		      adjust_anchor (g->f.f4.mark_anchor, face, g, &mark_x, &mark_y);
+		    xoff = (base_x - prev_width) - mark_x;
+		    yoff = base_y - mark_y;
+		  }
+		}
+	      if (i + 1 == gstring.used
+		  || gstring.glyphs[i + 1].glyph_id
+		  || ! gstring.glyphs[i + 1].positioning_type)
+		break;
+	      i++, g++;
+	    }
 	}
-	  
+
       XCopyArea (display, bmp->pixmap, seq_pixmap, gc_or,
 		 glyph_x + bmp->x, glyph_y + bmp->y, bmp->width, bmp->height,
 		 x + bmp->x + xoff, glyph_y + bmp->y - yoff);
@@ -467,28 +526,50 @@ update_render_area ()
   XFillRectangle (display, raw_pixmap, gc, 0, 0, render_width, render_height);
   for (i = 0, x = glyph_x; i < glyph_rec.n_glyphs; i++)
     {
-      BitmapRec *bmp = bitmap + glyph_rec.glyphs[i];
-      char buf[5];
+      if (glyph_rec.glyphs[i] >= 0)
+	{
+	  BitmapRec *bmp = bitmap + glyph_rec.glyphs[i];
+	  char buf[5];
 
-      XCopyArea (display, bmp->pixmap, raw_pixmap, gc,
-		 0, 0, glyph_width, glyph_height,
-		 (glyph_width + 4) * i + 1, 1);
-      XDrawRectangle (display, raw_pixmap, gc_set,
-		      (glyph_width + 4) * i, 0,
-		      glyph_width + 1, glyph_height + 1);
-      XDrawLine (display, raw_pixmap, gc_set,
-		 (glyph_width + 4) * i + 1 + glyph_x, 1,
-		 (glyph_width + 4) * i + 1 + glyph_x, glyph_height + 1);
-      XDrawLine (display, raw_pixmap, gc_set,
-		 (glyph_width + 4) * i + 1 + glyph_x + bmp->advance, 1,
-		 (glyph_width + 4) * i + 1 + glyph_x + bmp->advance,
-		 glyph_height + 1);
+	  XCopyArea (display, bmp->pixmap, raw_pixmap, gc,
+		     0, 0, glyph_width, glyph_height,
+		     (glyph_width + 4) * i + 1, 1);
+	  XDrawRectangle (display, raw_pixmap, gc_set,
+			  (glyph_width + 4) * i, 0,
+			  glyph_width + 1, glyph_height + 1);
+	  XDrawLine (display, raw_pixmap, gc_set,
+		     (glyph_width + 4) * i + 1 + glyph_x, 1,
+		     (glyph_width + 4) * i + 1 + glyph_x, glyph_height + 1);
+	  XDrawLine (display, raw_pixmap, gc_set,
+		     (glyph_width + 4) * i + 1 + glyph_x + bmp->advance, 1,
+		     (glyph_width + 4) * i + 1 + glyph_x + bmp->advance,
+		     glyph_height + 1);
 
-      sprintf (buf, "%04X", glyph_rec.codes[i]);
-      XDrawString (display, raw_pixmap, gc_inv, 
-		   (glyph_width + 1) * i + 1
-		   + (glyph_width - XTextWidth (font, buf, 4)) / 2,
-		   glyph_height + 2 + FONT_HEIGHT, buf, 4);
+	  sprintf (buf, "%04X", glyph_rec.codes[i]);
+	  XDrawString (display, raw_pixmap, gc_inv, 
+		       (glyph_width + 4) * i + 1
+		       + (glyph_width - XTextWidth (font, buf, 4)) / 2,
+		       glyph_height + 2 + FONT_HEIGHT, buf, 4);
+	}
+      else
+	{
+	  /* Variation Selector */
+	  int idx = - glyph_rec.glyphs[i];
+	  char buf[4];
+
+	  sprintf (buf, "%03d", idx);
+	  XDrawRectangle (display, raw_pixmap, gc_set,
+			  (glyph_width + 4) * i, 0,
+			  glyph_width + 1, glyph_height + 1);
+	  XDrawString (display, raw_pixmap, gc_set,
+		       (glyph_width + 4) * i + 1
+		       + (glyph_width - XTextWidth (font, "VS", 2)) / 2,
+		       1 + glyph_height / 2, "VS", 2);
+	  XDrawString (display, raw_pixmap, gc_set,
+		       (glyph_width + 4) * i + 1
+		       + (glyph_width - XTextWidth (font, buf, 3)) / 2,
+		       1 + glyph_height / 2 + FONT_ASCENT, buf, 3);
+	}
     }
   XtSetArg (arg[0], XtNbitmap, raw_pixmap);
   XtSetValues (raw_image, arg, 1);
@@ -640,18 +721,21 @@ void
 GlyphProc (Widget w, XtPointer client_data, XtPointer call_data)
 {
   int old_glyph_index = glyph_index;
+  int data;
 
-  if ((int) client_data == -3 && glyph_index > 0)
+  CAST_FROM_XTPOINTER (int, client_data, data);
+
+  if (data == -3 && glyph_index > 0)
     glyph_index = 0;
-  else if ((int) client_data == -2 && glyph_index > 0)
+  else if (data == -2 && glyph_index > 0)
     glyph_index = (glyph_index - 1) & 0x1FF000;
-  else if ((int) client_data == -1 && glyph_index > 0)
+  else if (data == -1 && glyph_index > 0)
     glyph_index -= 0x80;
-  else if ((int) client_data == 1 && glyph_index < 0x10FF80)
+  else if (data == 1 && glyph_index < 0x10FF80)
     glyph_index += 0x80;
-  else if ((int) client_data == 2 && glyph_index < 0x10F000)
+  else if (data == 2 && glyph_index < 0x10F000)
     glyph_index = (glyph_index + 0x1000) & 0x1FF000;
-  else if ((int) client_data == 3 && glyph_index < 0x10F000)
+  else if (data == 3 && glyph_index < 0x10F000)
     glyph_index = 0x10FF80;
   if (glyph_index != old_glyph_index)
     update_glyph_area ();
@@ -660,22 +744,60 @@ GlyphProc (Widget w, XtPointer client_data, XtPointer call_data)
 void
 CharmapProc (Widget w, XtPointer client_data, XtPointer call_data)
 {
-  if (charmap_index == (int) client_data)
+  int data;
+
+  CAST_FROM_XTPOINTER (int, client_data, data);
+
+  if (charmap_index == data)
     return;
-  charmap_index = (int) client_data;
+  charmap_index = data;
   if (charmap_index >= 0)
     FT_Set_Charmap (face, face->charmaps[charmap_index]);
   update_glyph_area ();
 }
 
+void
+UVSProc (Widget w, XtPointer client_data, XtPointer call_data)
+{
+  unsigned idx;
+  int selector;
+  OTF_VariationSelectorRecord *record;
+  int i;
+
+  CAST_FROM_XTPOINTER (unsigned, client_data, idx);
+  selector = uvs[idx].c;
+
+  if (glyph_rec.n_glyphs >= 64)
+    return;
+  for (i = 0; i < sub14->nRecords; i++)
+    {
+      record = sub14->Records + i;
+      if (record->varSelector == selector)
+	break;
+    }
+  if (i < sub14->nRecords)
+    {
+      if (glyph_rec.n_glyphs > 0
+	  && glyph_rec.glyphs[glyph_rec.n_glyphs - 1] < 0)
+	glyph_rec.n_glyphs--;
+      glyph_rec.codes[glyph_rec.n_glyphs] = selector;
+      glyph_rec.glyphs[glyph_rec.n_glyphs++] = - idx - 1;
+      update_render_area ();
+    }
+}
+
 void
 RenderProc (Widget w, XtPointer client_data, XtPointer call_data)
 {
-  if ((int) client_data < 0)
+  int data;
+
+  CAST_FROM_XTPOINTER (int, client_data, data);
+
+  if (data < 0)
     {
       if (glyph_rec.n_glyphs > 0)
 	{
-	  if ((int) client_data == -2)
+	  if (data == -2)
 	    glyph_rec.n_glyphs--;
 	  else
 	    glyph_rec.n_glyphs = 0;
@@ -684,14 +806,16 @@ RenderProc (Widget w, XtPointer client_data, XtPointer call_data)
     }
   else if (glyph_rec.n_glyphs < 64)
     {
-      int index = glyph_index + (int) client_data;
+      int index = glyph_index + data;
 
       if (charmap_index >= 0)
 	index = FT_Get_Char_Index (face, (FT_ULong) index);
       if (bitmap[index].pixmap)
 	{
-	  glyph_rec.codes[glyph_rec.n_glyphs] = glyph_index + (int) client_data;
+	  glyph_rec.codes[glyph_rec.n_glyphs] = glyph_index + data;
 	  glyph_rec.glyphs[glyph_rec.n_glyphs++] = index;
+	  if (otf)
+	    update_uvs_area (glyph_index + data);
 	  update_render_area ();
 	}
     }
@@ -1132,7 +1256,41 @@ create_widgets ()
   XtSetArg (arg[6], XtNdefaultDistance, 0);
   glyph_area = XtCreateManagedWidget ("glyph-area", formWidgetClass,
 				      frame, arg, 7);
+
   XtSetArg (arg[5], XtNfromVert, glyph_area);
+  if (sub14)
+    {
+      Arg arg2[3];
+
+      XtSetArg (arg[6], XtNorientation, XtorientHorizontal);
+      uvs_area = XtCreateManagedWidget ("uvs-area", boxWidgetClass,
+					frame, arg, 7);
+      XtSetArg (arg2[0], XtNborderWidth, 0);
+      XtSetArg (arg2[1], XtNlabel, "Variation Selector: ");
+      uvs_label = XtCreateManagedWidget ("uvs-label", labelWidgetClass,
+					 uvs_area, arg2, 2);
+      XtSetArg (arg2[0], XtNborderWidth, 1);
+      for (i = 0; i < sub14->nRecords; i++)
+	{
+	  OTF_VariationSelectorRecord *record = sub14->Records + i;
+	  unsigned selector = record->varSelector;
+	  unsigned idx;
+	  char lbl[4];
+	  
+	  idx = (selector <= 0xFE0F ? selector - 0xFE00
+		 : selector - 0xE0100 + 16);
+	  if (uvs[idx].c)
+	    continue;
+	  uvs[idx].c = selector;
+	  sprintf (lbl, "%03d", idx + 1);
+	  XtSetArg (arg2[1], XtNlabel, lbl);
+	  XtSetArg (arg2[2], XtNsensitive, False);
+	  uvs[idx].w = XtCreateManagedWidget ("lbl", commandWidgetClass,
+						   uvs_area, arg2, 3);
+	  XtAddCallbackWithCast (unsigned, uvs[idx].w, UVSProc, idx);
+	}
+      XtSetArg (arg[5], XtNfromVert, uvs_area);
+    }
   render_area = XtCreateManagedWidget ("render-area", formWidgetClass,
 				       frame, arg, 6);
 
@@ -1160,7 +1318,7 @@ create_widgets ()
       charmap[i + 1] = XtCreateManagedWidget (charmap_rec[i + 1].name,
 					      toggleWidgetClass,
 					      command_area, arg, 1);
-      XtAddCallback (charmap[i + 1], XtNcallback, CharmapProc, (XtPointer) i);
+      XtAddCallbackWithCast (int, charmap[i + 1], CharmapProc, i);
     }
 
   XtSetArg (arg[0], XtNlabel, " |< (f)");
@@ -1234,7 +1392,7 @@ create_widgets ()
 	  XtSetArg (arg[n], XtNbitmap, none_pixmap), n++;
 	  glyph[k] = XtCreateManagedWidget ("glyph", commandWidgetClass,
 					    glyph_area, arg, n);
-	  XtAddCallback (glyph[k], XtNcallback, RenderProc, (XtPointer) k);
+	  XtAddCallbackWithCast (int, glyph[k], RenderProc, k);
 	}
       w = head;
     }
@@ -1427,12 +1585,24 @@ main (int argc, char **argv)
       || strstr (filename, ".OTF"))
     {
       otf = OTF_open_ft_face (face);
-      if (! otf
-	  || OTF_get_table (otf, "head") < 0
-	  || OTF_get_table (otf, "cmap") < 0
-	  || (OTF_check_table (otf, "GSUB") < 0
-	      && OTF_check_table (otf, "GPOS") < 0))
-	otf = NULL;
+      if (otf)
+	{
+	  if (OTF_get_table (otf, "head") < 0
+	      || OTF_get_table (otf, "cmap") < 0
+	      || (OTF_check_table (otf, "GSUB") < 0
+		  && OTF_check_table (otf, "GPOS") < 0))
+	    {
+	      OTF_close (otf);
+	      otf = NULL;
+	    }
+	}
+      if (otf)
+	for (i = 0; i < otf->cmap->numTables; i++)
+	  if (otf->cmap->EncodingRecord[i].subtable.format == 14)
+	    {
+	      sub14 = otf->cmap->EncodingRecord[i].subtable.f.f14;
+	      break;
+	    }
     }
 
   {
@@ -1471,9 +1641,11 @@ main (int argc, char **argv)
     XGCValues values;
 
     gc = XCreateGC (display, none_pixmap, (unsigned long) 0, NULL);
+    XSetFont (display, gc, font->fid);
     values.function = GXset;
     values.line_width = 1;
     gc_set = XCreateGC (display, none_pixmap, valuemask, &values);
+    XSetFont (display, gc_set, font->fid);
     values.function = GXor;
     gc_or = XCreateGC (display, none_pixmap, valuemask, &values);
     values.function = GXcopyInverted;
