@@ -1,6 +1,6 @@
 /* otf.h -- Header file for libotf (OpenType font library).
 
-Copyright (C) 2003, 2004, 2005, 2006, 2007
+Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009
   National Institute of Advanced Industrial Science and Technology (AIST)
   Registration Number H15PRO167
 
@@ -29,14 +29,14 @@ extern "C" {
 #endif
 
 /* Version name of this library.  */
-#define LIBOTF_VERSION "0.9.6"
+#define LIBOTF_VERSION "0.9.9"
 
 /* Major version number.  */
 #define LIBOTF_MAJOR_VERSION 0
 /* Minor version number.  */
 #define LIBOTF_MINOR_VERSION 9
 /* Release (i.e. patch level) number.  */
-#define LIBOTF_RELEASE_NUMBER 6
+#define LIBOTF_RELEASE_NUMBER 9
 
 /***
     Table of contents:
@@ -165,7 +165,7 @@ typedef struct
   unsigned endCount;
   int idDelta;
   unsigned idRangeOffset;
-} OTF_cmapSegument;
+} OTF_cmapSegment;
 
 typedef struct
 {
@@ -173,7 +173,7 @@ typedef struct
   unsigned searchRange;
   unsigned entrySelector;
   unsigned rangeShift;
-  OTF_cmapSegument *segments;
+  OTF_cmapSegment *segments;
   int GlyphCount;
   unsigned *glyphIdArray;
 } OTF_EncodingSubtable4;
@@ -212,6 +212,37 @@ typedef struct
   OTF_cmapGroup *Groups;
 } OTF_EncodingSubtable12;
 
+typedef struct
+{
+  unsigned unicodeValue;
+  unsigned short glyphID;
+} OTF_UVSMapping;
+
+typedef struct
+{
+  unsigned startUnicodeValue;
+  unsigned short additionalCount;
+} OTF_UnicodeValueRange;
+
+typedef struct
+{
+  unsigned varSelector;
+  unsigned defaultUVSOffset;
+  unsigned nonDefaultUVSOffset;
+  /* DefaultUVS */
+  unsigned numUnicodeValueRanges;
+  OTF_UnicodeValueRange *unicodeValueRanges;
+  /* NonDefaultUVS */
+  unsigned numUVSMappings;
+  OTF_UVSMapping *uvsMappings;
+} OTF_VariationSelectorRecord;
+
+typedef struct
+{
+  unsigned nRecords;
+  OTF_VariationSelectorRecord *Records;
+} OTF_EncodingSubtable14;
+
 typedef struct
 {
   unsigned format;
@@ -225,6 +256,7 @@ typedef struct
     OTF_EncodingSubtable8 *f8;
     OTF_EncodingSubtable10 *f10;
     OTF_EncodingSubtable12 *f12;
+    OTF_EncodingSubtable14 *f14;
   }f;
 } OTF_EncodingSubtable;
 
@@ -1186,7 +1218,14 @@ typedef struct
   int c;
 
   /* Glyph ID of the glyph.  If the value is 0, the library gets a
-     correct value from the above character code via cmap.  */
+     correct value from the above character code via cmap if such a
+     glyph is available in the font.
+
+     The function OTF_drive_gpos2 may insert a glyph whose glyph_id is
+     0 but positioning_type is positive.  It is not an actual glyph
+     but just contains positioning information that should be
+     accumulated to the positioning information of the previous
+     glyphs.  */
   OTF_GlyphID glyph_id;
 
   /* GlyphClass of the glyph.  The value is extracted from the GDEF
@@ -1258,7 +1297,7 @@ typedef struct
     Process glyph string by Unicode-based cmap table.
 
     The OTF_drive_cmap() function looks up a Unicode-based cmap table
-    of OpenType font $OTF, and setup the member <glyhph_id> of all
+    of OpenType font $OTF, and setup the member <glyph_id> of all
     glhphs in the glyph string $GSTRING if the value of the member is
     not zero.  */
 
@@ -1269,13 +1308,23 @@ extern int OTF_drive_cmap (OTF *otf, OTF_GlyphString *gstring);
 
     The OTF_drive_cmap2() function looks up a cmap table (whose
     Platform-ID is $PLATFORM_ID an Encoding-ID is $ENCODING_ID) of
-    OpenType font $OTF, and setup the member <glyhph_id> of all glhphs
+    OpenType font $OTF, and setup the member <glyph_id> of all glhphs
     in the glyph string $GSTRING if the value of the member is not
     zero.  */
 
 extern int OTF_drive_cmap2 (OTF *otf, OTF_GlyphString *gstring,
 			    int platform_id, int encoding_id);
 
+
+/***
+    Store variable glyphs of character C in the array CODE.  The array
+    size must be 256.  The Nth element of CODE is the glyph corresponding
+    to the variation selector (N + 1).  The return value is the number
+    of variation glyphs.  */
+
+extern int OTF_get_variation_glyphs (OTF *otf, int c, OTF_GlyphID code[256]);
+
+
 /*** (3-3) OTF_drive_gdef() */
 
 /***
@@ -1310,15 +1359,25 @@ extern int OTF_drive_gsub (OTF *otf, OTF_GlyphString *gstring,
 /***
     Process glyph string by GPOS table.
 
-    The OTF_drive_gdef() function looks up the GPOS table of $OTF of
+    The OTF_drive_gpos() function is deprecated.  Use
+    OTF_drive_gpos2() instread.  */
+
+extern int OTF_drive_gpos (OTF *otf, OTF_GlyphString *gstring,
+			   const char *script, const char *language,
+			   const char *features);
+
+/***
+    Process glyph string by GPOS table.
+
+    The OTF_drive_gpos2() function looks up the GPOS table of $OTF of
     OpenType font $OTF, and by using features the font has for script
     $SCRIPT and language system $LANGSYS, setup members
     <positioning_type> and <f> of all glhphs in the glyph string
     $GSTRING.  $FEATURES is a list of features to apply.  */
 
-extern int OTF_drive_gpos (OTF *otf, OTF_GlyphString *gstring,
-			   const char *script, const char *language,
-			   const char *features);
+extern int OTF_drive_gpos2 (OTF *otf, OTF_GlyphString *gstring,
+			    const char *script, const char *language,
+			    const char *features);
 
 /*** (3-6) OTF_drive_tables() */
 
