@@ -1,6 +1,6 @@
 /* otfopen.c -- OpenType font reader.
 
-Copyright (C) 2003, 2004, 2005, 2006, 2008
+Copyright (C) 2003, 2004, 2005, 2006, 2008, 2009
   National Institute of Advanced Industrial Science and Technology (AIST)
   Registration Number H15PRO167
 
@@ -91,7 +91,7 @@ typedef struct
 
 typedef long OTF_StreamState;
 
-OTF_Stream *
+static OTF_Stream *
 make_stream (const char *name)
 {
   OTF_Stream *stream;
@@ -105,7 +105,7 @@ make_stream (const char *name)
   return stream;
 }
 
-int
+static int
 setup_stream (OTF_Stream *stream, FILE *fp, long offset, int nbytes)
 {
   char *errfmt = "stream setup for %s";
@@ -131,7 +131,7 @@ setup_stream (OTF_Stream *stream, FILE *fp, long offset, int nbytes)
   return 0;
 }
 
-OTF_Stream *
+static OTF_Stream *
 make_stream_from_ft_face (FT_Face face, const char *name)
 {
   char *errfmt = "FT_Face stream creation for %s";
@@ -161,10 +161,11 @@ make_stream_from_ft_face (FT_Face face, const char *name)
   return stream;
 }
 
-void
+static void
 free_stream (OTF_Stream *stream)
 {
-  free (stream->buf);
+  if (stream->buf)
+    free (stream->buf);
   free (stream);
 }
 
@@ -199,6 +200,15 @@ free_stream (OTF_Stream *stream)
     (stream)->pos += 2;						\
   } while (0)
 
+#define READ_UINT24(stream, var)			\
+  do {							\
+    STREAM_CHECK_SIZE ((stream), 3);			\
+    (var) =  (((stream)->buf[(stream)->pos ] << 16)	\
+	      | ((stream)->buf[(stream)->pos + 1] << 8)	\
+	      | (stream)->buf[(stream)->pos + 2]);	\
+    (stream)->pos += 3;					\
+  } while (0)
+
 #define READ_ULONG(stream, var)				\
   do {							\
     STREAM_CHECK_SIZE ((stream), 4);			\
@@ -382,13 +392,14 @@ read_head_table (OTF *otf, OTF_TableInfo *table, enum OTF_ReaderFlag flag)
 static int
 read_name (OTF *otf, OTF_Stream *stream, OTF_NameRecord *rec)
 {
-  char *errfmt = "nameID (%d)";
+  char errfmt[256];
   int errret = -1;
   OTF_StreamState state;
   int ucs = 0;
   int ascii = 0;
   int i;
 
+  sprintf (errfmt, "nameID (%d)%%s", rec->nameID);
   if (rec->platformID == 0)
     ucs = (rec->encodingID <= 3) ? 2 : 4;
   else if (rec->platformID == 1 && rec->encodingID == 0)
@@ -398,7 +409,7 @@ read_name (OTF *otf, OTF_Stream *stream, OTF_NameRecord *rec)
 	   : rec->encodingID == 10 ? 4
 	   : 0);
 
-  OTF_MALLOC (rec->name, rec->length + 1, (void *) rec->nameID);
+  OTF_MALLOC (rec->name, rec->length + 1, "");
   SAVE_STREAM (stream, state);
   SEEK_STREAM (stream, stream->pos + rec->offset);
   READ_BYTES (stream, rec->name, rec->length);
@@ -495,6 +506,85 @@ read_name_table (OTF *otf, OTF_TableInfo *table, enum OTF_ReaderFlag flag)
 
 /*** (1-4) "cmap" table */
 
+static OTF_EncodingSubtable14 *
+read_cmap_uvs_table (OTF *otf, OTF_Stream *stream, OTF_Offset offset)
+{
+  OTF_EncodingSubtable14 *sub14;
+  char *errfmt = "cmap-uvs%s";
+  void *errret = NULL;
+  unsigned nRecords;
+  unsigned i,j;
+
+  OTF_MALLOC (sub14, 1, " (EncodingSubtable14)");
+  READ_ULONG (stream, nRecords);
+  sub14->nRecords = nRecords;
+  OTF_MALLOC (sub14->Records, nRecords, "(EncodingSubtable14-Records)");
+  for (i = 0; i < sub14->nRecords; i++) 
+    {
+      unsigned varSelector=0, defaultUVSOffset, nonDefaultUVSOffset;
+
+      READ_UINT24 (stream, varSelector);
+      sub14->Records[i].varSelector = varSelector;
+      READ_ULONG (stream, defaultUVSOffset);
+      sub14->Records[i].defaultUVSOffset = defaultUVSOffset;
+      READ_ULONG (stream, nonDefaultUVSOffset);
+      sub14->Records[i].nonDefaultUVSOffset = nonDefaultUVSOffset;
+    }
+  for (i = 0; i < sub14->nRecords; i++) 
+    {
+      OTF_VariationSelectorRecord *record = &sub14->Records[i];
+      unsigned defaultUVSOffset = record->defaultUVSOffset;
+      unsigned nonDefaultUVSOffset = record->nonDefaultUVSOffset;
+
+      if (defaultUVSOffset) 
+	{
+	  unsigned numUnicodeValueRanges;
+
+	  SEEK_STREAM (stream, offset+defaultUVSOffset);
+	  READ_ULONG (stream, numUnicodeValueRanges);
+	  record->numUnicodeValueRanges = numUnicodeValueRanges;
+	  OTF_MALLOC (record->unicodeValueRanges, 
+		      numUnicodeValueRanges,
+		      "(EncodingSubtable14-Records-unicodeValueRanges)");
+	  for (j = 0; j < numUnicodeValueRanges; j++)
+	    {
+	      OTF_UnicodeValueRange *unicodeValueRange 
+		= &record->unicodeValueRanges[j];
+	      unsigned startUnicodeValue;
+	      char additionalCount;
+
+	      READ_UINT24 (stream, startUnicodeValue);
+	      unicodeValueRange->startUnicodeValue=startUnicodeValue;
+	      READ_BYTES (stream, &additionalCount, 1);
+	      unicodeValueRange->additionalCount
+		= (unsigned short) additionalCount;
+	    }
+	}
+      if (nonDefaultUVSOffset) 
+	{
+	  unsigned numUVSMappings;
+
+	  SEEK_STREAM (stream, offset+nonDefaultUVSOffset);
+	  READ_ULONG (stream, numUVSMappings);
+	  record->numUVSMappings = numUVSMappings;
+	  OTF_MALLOC (record->uvsMappings, numUVSMappings,
+		      "(EncodingSubtable14-Records-uvsMappings)");
+	  for (j = 0; j < numUVSMappings; j++)
+	    {
+	      OTF_UVSMapping *uvsMapping = &record->uvsMappings[j];
+	      unsigned unicodeValue;
+	      unsigned short glyphID;
+
+	      READ_UINT24 (stream, unicodeValue);
+	      uvsMapping->unicodeValue = unicodeValue;
+	      READ_USHORT (stream, glyphID);
+	      uvsMapping->glyphID = glyphID;
+	    }
+	}
+    }
+  return sub14;
+}
+
 static void *
 read_cmap_table (OTF *otf, OTF_TableInfo *table, enum OTF_ReaderFlag flag)
 {
@@ -540,15 +630,23 @@ read_cmap_table (OTF *otf, OTF_TableInfo *table, enum OTF_ReaderFlag flag)
       SEEK_STREAM (stream, cmap->EncodingRecord[i].offset);
       READ_USHORT (stream, format);
       cmap->EncodingRecord[i].subtable.format = format;
-      READ_USHORT (stream, cmap->EncodingRecord[i].subtable.length);
-      if (format == 8 || format == 10 || format == 12)
+      if (format == 14)
 	{
-	  READ_ULONG (stream, cmap->EncodingRecord[i].subtable.length);
-	  READ_ULONG (stream, cmap->EncodingRecord[i].subtable.language);
+	  READ_ULONG (stream, cmap->EncodingRecord[i].subtable.length);	    
+	  cmap->EncodingRecord[i].subtable.language = 0;
 	}
       else
 	{
-	  READ_USHORT (stream, cmap->EncodingRecord[i].subtable.language);
+	  READ_USHORT (stream, cmap->EncodingRecord[i].subtable.length);
+	  if (format == 8 || format == 10 || format == 12)
+	    {
+	      READ_ULONG (stream, cmap->EncodingRecord[i].subtable.length);
+	      READ_ULONG (stream, cmap->EncodingRecord[i].subtable.language);
+	    }
+	  else
+	    {
+	      READ_USHORT (stream, cmap->EncodingRecord[i].subtable.language);
+	    }
 	}
       switch (format)
 	{
@@ -705,6 +803,14 @@ read_cmap_table (OTF *otf, OTF_TableInfo *table, enum OTF_ReaderFlag flag)
 	  }
 	  break;
 
+	case 14:
+	  {
+	    cmap->EncodingRecord[i].subtable.f.f14
+	      = read_cmap_uvs_table (otf, stream,
+				     cmap->EncodingRecord[i].offset);
+	    break;
+	  }
+
 	default:
 	  OTF_ERROR (OTF_ERROR_TABLE, " (invalid Subtable format)");
 	}
@@ -725,7 +831,7 @@ read_cmap_table (OTF *otf, OTF_TableInfo *table, enum OTF_ReaderFlag flag)
 
 	    for (i = 0; i < segCount; i++)
 	      {
-		OTF_cmapSegument *seg = sub4->segments + i;
+		OTF_cmapSegment *seg = sub4->segments + i;
 		int c;
 
 		if (seg->idRangeOffset == 0xFFFF)
@@ -1106,22 +1212,27 @@ read_gdef_table (OTF *otf, OTF_TableInfo *table, enum OTF_ReaderFlag flag)
   OTF_GDEF *gdef;
 
   OTF_CALLOC (gdef, 1, "");
-  read_gdef_header (stream, (OTF_GDEFHeader *) &gdef->header);
-  if (gdef->header.GlyphClassDef)
+  if (stream->buf)
     {
-      gdef->glyph_class_def.offset = gdef->header.GlyphClassDef;
-      read_class_def_without_offset (otf, stream, &gdef->glyph_class_def);
-    }
-  if (gdef->header.AttachList)
-    read_attach_list (otf, stream, gdef->header.AttachList,
-		      &gdef->attach_list);
-  if (gdef->header.LigCaretList)
-    read_lig_caret_list (otf, stream, gdef->header.LigCaretList,
-			 &gdef->lig_caret_list);
-  if (gdef->header.MarkAttachClassDef)
-    {
-      gdef->mark_attach_class_def.offset = gdef->header.MarkAttachClassDef;
-      read_class_def_without_offset (otf, stream, &gdef->mark_attach_class_def);
+      read_gdef_header (stream, (OTF_GDEFHeader *) &gdef->header);
+      if (gdef->header.GlyphClassDef)
+	{
+	  gdef->glyph_class_def.offset = gdef->header.GlyphClassDef;
+	  read_class_def_without_offset (otf, stream,
+					 &gdef->glyph_class_def);
+	}
+      if (gdef->header.AttachList)
+	read_attach_list (otf, stream, gdef->header.AttachList,
+			  &gdef->attach_list);
+      if (gdef->header.LigCaretList)
+	read_lig_caret_list (otf, stream, gdef->header.LigCaretList,
+			     &gdef->lig_caret_list);
+      if (gdef->header.MarkAttachClassDef)
+	{
+	  gdef->mark_attach_class_def.offset = gdef->header.MarkAttachClassDef;
+	  read_class_def_without_offset (otf, stream,
+					 &gdef->mark_attach_class_def);
+	}
     }
 
   *table->address = gdef;
@@ -2611,7 +2722,7 @@ read_jstf_table (OTF_Stream *stream, long offset)
 
 /*** (1-11) Structure for OTF */
 
-int
+static int
 read_offset_table (OTF *otf, OTF_Stream *stream, OTF_OffsetTable *table)
 {
   int errret = -1;
@@ -2742,6 +2853,10 @@ read_header_part (OTF *otf, FILE *fp, FT_Face face)
 	internal_data->table_info[OTF_TABLE_TYPE_GPOS].stream = stream;
     }
 
+  if (! internal_data->table_info[OTF_TABLE_TYPE_GDEF].stream)
+    /* We can simulate the GDEF table.  */
+    internal_data->table_info[OTF_TABLE_TYPE_GDEF].stream
+      = make_stream ("GDEF");
   return 0;
 }
 
